def bressenham(canvas, line, val):
  """
  based on 'optimized' version at http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
  """
  (x0,y0,x1,y1) = line
  
  steep = abs(y1 - y0) > abs(x1 - x0)
  if steep:
    x0, y0 = y0, x0  
    x1, y1 = y1, x1

  if x0 > x1:
    x0, x1 = x1, x0
    y0, y1 = y1, y0

  x0,x1=(x0-x0,x1-x0) #canvas is assumed to be a tight bounding box; 
  y0,y1=(y0-y0,y1-y0) #  offset the points relative to the top-left

  deltax = x1 - x0
  deltay = abs(y1 - y0)
  error = 0
  ystep = 1 if y0 < y1 else -1
  
  y = y0
  for x in range(x0, x1): #line only includes first endpoint
    #import pdb; pdb.set_trace()
    if steep:
      if x >= canvas.shape[1]: break
      if y >= canvas.shape[0]: break
      canvas[y,x]=val
    else:
      if x >= canvas.shape[0]: break
      if y >= canvas.shape[1]: break
      canvas[x,y]=val

    error = error - deltay
    if (error < 0):
      y = y + ystep
      error = error + deltax



"""Core abstract rendering abstractions.  This includes the main drivers of execution
   and the base clases for shared data representations.
"""

import re
import sys
import os
import numpy as np
import ctypes
from fast_project import _projectRects
import geometry
import glyphset



try:
  from numba import autojit
except ImportError:
  print "Error loading numba."
  autojit = lambda f: f

_lib = ctypes.CDLL(os.path.join(os.path.dirname(__file__), 'transform.so'))

############################  Core System ####################
def glyphAggregates(points, shapeCode, val, default):
  """Create a set of aggregates fo a single glyph.  The set of aggregates will be
     tight to the bound box of the shape but may not be completely filled 
     (thus the need for both 'val' and 'default').

     * points -- Points that define the glyph
     * shapeCode -- Code that indicates how to interpret the points
     * val -- Value to place in bins that are hit by the shape
     * default -- Value to place in bins not hit by the shape
  """

  def scalar(array, val): array.fill(val)
  def nparray(array,val): array[:] = val
  
 
  if type(val) == np.ndarray:
    fill = nparray 
    extShape = val.shape
  else:
    fill = scalar 
    extShape = ()

  #TODO: These are selectors...rename and move this somewhere else
  if shapeCode == glyphset.ShapeCodes.POINT:
    array = np.copy(val) ##TODO: Not sure this is always an array...should verify that
  elif shapeCode == glyphset.ShapeCodes.RECT:
    array = np.empty((points[2]-points[0],points[3]-points[1])+extShape, dtype=np.int32)
    fill(array, val)
  elif shapeCode == glyphset.ShapeCodes.LINE:
    array = np.empty((points[2]-points[0],points[3]-points[1])+extShape, dtype=np.int32)
    fill(array, default)
    geometry.bressenham(array, points, val)

  return array


############ Core process functions #################

def render(glyphs, info, aggregator, shader, screen,ivt):
  """
  Render a set of glyphs under the specified condition to the described canvas.

  * glyphs -- Glyphs to render
  * info -- For each glyph, what is the piece of information that will be aggregated
  * aggregator -- Combines a set of info values into a single aggregate value
  * shader -- Converts aggregates to other aggregates (often colors)
  * screen -- (width,height) of the canvas
  * ivt -- INVERSE view transform (converts pixels to canvas space)

  """
  projected = project(glyphs, ivt.inverse())
  aggregates = aggregate(projected, info, aggregator, screen)
  shaded = shade(aggregates, shader)
  return shaded


def project(glyphs, viewxform):
  """Project the points found in the glyphset according to the view transform.

     * viewxform -- convert canvas space to pixel space
     * glyphs -- set of glyphs (represented as [x,y,w,h,...]
  """
  points = glyphs.points()
  out = np.empty_like(points, dtype=np.int32)
  _projectRects(viewxform.asarray(), points, out)
  
  #Ensure visilibity, make sure w/h are always at least one
  #TODO: There is probably a more numpy-ish way to do this...(and it might not be needed for Shapecode.POINT)
  for i in xrange(0,out.shape[0]):
    if out[i,0] == out[i,2]: out[i,2] += 1
    if out[i,1] == out[i,3]: out[i,3] += 1

  return glyphset.Glyphset(out, glyphs.data(), glyphset.Literals(glyphs.shaper.code))

def aggregate(glyphs, info, aggregator, screen):
    (width, height) = screen

    infos = [info(point, data) for point, data in zip(glyphs.points(), glyphs.data())] #TODO: vectorize
    aggregates = aggregator.allocate(width, height, glyphs, infos)
    for idx, points in enumerate(glyphs.points()):
      aggregator.combine(aggregates, points, glyphs.shaper.code, infos[idx])
    return aggregates


#TODO: Add specialization here.  Take a 3rd argument 'specailizer'  if ommited, just use aggregates
def shade(aggregates, shader):
   """Convert a set of aggregate into another set of aggregates
      according to some data shader.  Many common cases, the result
      aggregates is an image, but it does not need to be.

      NOTE:  This is currently a rather simple function.  It is included now
      as an extension point.

      * aggregates -- input aggregaets
      * shader -- data shader used in the conversion
   """
   return shader.shade(aggregates)


class Aggregator(object):
  out_type = None
  in_type = None
  identity=None
  
  def allocate(self, width, height, glyphset, infos):
    pass

  def combine(self, existing, points, shapecode, val):
    """
    * existing - out_type numpy array, aggregate values for all glyphs seen thus far
    * points - points that define a shape
    * shapecode - Code that determines how points are interpreted
    * val -- Info value associated with the current set of points
    """
    pass

  def rollup(*vals):
    """
    Combine multiple sets of aggregates.

    * *vals - list of numpy arrays with type out_type
    """
    pass


#TODO: Add specialization to Shaders....
class Shader(object):
  def makegrid(self, grid):
    """Create an output grid.  
       Default implementation creates one of the same width/height of the input
       suitable for colors (dept 4, unit8).
    """
    (width, height) = grid.shape[0], grid.shape[1]
    return np.ndarray((width, height, 4), dtype=np.uint8)

  def shade(self, grid):
    """Execute the actual data shader operation."""
    raise NotImplementedError
  
  def __add__(self, other): 
    """Extend this shader by executing another in sequence."""
    if (not isinstance(other, Shader)): 
        raise TypeError("Can only extend with a shader.  Received a " + str(type(other)))
    return Seq(self, other) 


class Seq(Shader):
  """Shader that does a sequence of other shaders."""
     
  def __init__(self, *args):
    self._parts = args

  def makegrid(self, grid):
    for t in self._parts:
      grid = t.makegrid(grid)
    return grid

  def shade(self, grid):
    for t in self._parts:
      grid = t.shade(grid)
    return grid

  def __add__(self, other):
    if (other is None) : return self
    if (not isinstance(other, Shader)): 
        raise TypeError("Can only extend shader with another shader.  Received a " + str(type(other)))
    return Seq(list(self._parts) + other) 


class PixelAggregator(Aggregator):
  def __init__(self, pixelfunc):
    self.pixelfunc = pixelfunc

  def aggregate(self, grid):
      outgrid = np.empty_like(self._projected, dtype=np.int32)
      #outgrid = np.empty_like(self._projected, dtype=aggregator.out_dtype)
      outgrid.ravel()[:] = map(lambda ids: self.pixelfunc(self._glyphset, ids), self._projected.flat)


class PixelShader(Shader):
  """Data shader that does non-vectorized per-pixel shading."""

  def __init__(self, pixelfunc, prefunc):
    self.pixelfunc = pixelfunc
    self.prefunc = prefunc
  
  def _pre(self, grid):
    """Executed exactly once before pixelfunc is called on any cell."""
    pass

  def pixelfunc(grid, x, y):
    """Override this method.  It will be called for each pixel in the grid."""
    raise NotImplementedError

  def shade(self, grid):
    outgrid = self.makegrid(grid)
    self._pre(grid)
    (width,height) = grid.shape
    for x in xrange(0, width):
      for y in xrange(0, height):
        outgrid[x,y] = self.pixelfunc(grid, x, y)

    return outgrid

###############################  Graphics Components ###############

class AffineTransform(list):
  def __init__(self, tx, ty, sx, sy):
    list.__init__(self, [tx,ty,sx,sy])
    self.tx=tx
    self.ty=ty
    self.sx=sx
    self.sy=sy

  def trans(self, x, y):
    """Transform a passed point."""
    x = self.sx * x + self.tx
    y = self.sy * y + self.ty
    return (x, y)

  def transform(self, glyph):
    """Transform a passed glyph (somethign with x,y,w,h)"""
    (p1x,p1y) = self.trans(glyph.x, glyph.y)
    (p2x,p2y) = self.trans(glyph.x+glyph.width, glyph.y+glyph.height)
    w = p2x-p1x
    h = p2y-p1y
    return Glyph(p1x, p1y, w, h, glyph.props)

  def asarray(self): return np.array(self)

  def inverse(self):
    return AffineTransform(-self.tx/self.sx, -self.ty/self.sy, 1/self.sx, 1/self.sy)

class Color(list):
  def __init__(self,r,g,b,a):
    list.__init__(self,[r,g,b,a])
    self.r=r
    self.g=g
    self.b=b
    self.a=a

  def asarray(self): return np.array(self, dtype=np.uint8)

class Glyph(list):
  def __init__(self,x,y,w,h,*props):
    fl = [x,y,w,h]
    fl.extend(props)
    list.__init__(self,fl)
    self.x=x
    self.y=y
    self.width=w
    self.height=h
    self.props=props

  def asarray(self): return np.array(self)


############################  Support functions ####################
class EmptyList(object):
  def __getitem__(self, idx): 
    return None

#Does the glyph contain any part of the pixel?
def contains(px, glyph):
  return (px.x+px.w > glyph.x   #Really is >= if using "left/top is in, right/bottom is out" convention
      and px.y + px.h > glyph.y #Really is >= if using "left/top is in, right/bottom is out" convention
      and px.x < glyph.x + glyph.width
      and px.y < glyph.y + glyph.height)

def containing(px, glyphs):
  items = []
  for g in glyphs:
    if contains(px, g): 
      items.append(g)
      
  return items

def zoom_fit(screen, bounds, balanced=True):
  """What affine transform will zoom-fit the given items?
     screen: (w,h) of the viewing region
     bounds: (x,y,w,h) of the items to fit
     balance: Should the x and y scales match?
     returns: AffineTransform object
  """
  (sw,sh) = screen
  (gx,gy,gw,gh) = bounds
  x_scale = gw/float(sw)
  y_scale = gh/float(sh)
  if (balanced):
    x_scale = max(x_scale, y_scale)
    y_scale = x_scale
  return AffineTransform(gx,gy,x_scale, y_scale)


def load_csv(filename, skip, xc,yc,vc,width,height):
  source = open(filename, 'r')
  glyphs = []
  data = []
  
  for i in range(0, skip):
    source.readline()

  for line in source:
    line = re.split("\s*,\s*", line)
    x = float(line[xc].strip())
    y = float(line[yc].strip())
    v = float(line[vc].strip()) if vc >=0 else 1 
    g = Glyph(x,y,width,height)
    glyphs.append(g)
    data.append(v)

  source.close()
  return glyphset.Glyphset(glyphs,data, glyphset.Literals(glyphset.ShapeCodes.RECT))

def main():
  ##Abstract rendering function implementation modules (for demo purposes only)
  import numeric
  import infos

  source = sys.argv[1]
  skip = int(sys.argv[2])
  xc = int(sys.argv[3])
  yc = int(sys.argv[4])
  vc = int(sys.argv[5])
  size = float(sys.argv[6])
  glyphs = load_csv(source,skip,xc,yc,vc,size,size)

  screen=(10,10)
  ivt = zoom_fit(screen,glyphs.bounds())

  image = render(glyphs, 
                 infos.id(),
                 numeric.Count(), 
                 numeric.AbsSegment(Color(0,0,0,0), Color(255,255,255,255), .5),
                 screen, 
                 ivt)

  print image


if __name__ == "__main__":
    main()



import core

class Id(core.Shader):
  """Return the input unchanged.  
     This DOES NOT make a copy.  
     It is usually used a zero-cost placeholder.
  """
  def shade(self, grid):
    return grid 




import numpy as np

def enum(**enums): return type('Enum', (), enums)
ShapeCodes = enum(POINT=0, LINE=1, RECT=2)  ##TODO: Add shapecodes: CIRCLE, POINT_RECT, POINT_CIRCLE, etc

class Glyphset(object):
  """shaper + shape-params + associated data ==> Glyphset

     fields:
        _points : Points held by this glyphset
        _ data : Data associated with the points (basis for 'info' values).  _points[x] should associate with data[x]
        shapecode: Shapecode that tells how to interpret _points

  """
  _points = None
  _data = None
  shaper = None

  def __init__(self, points, data, shaper, colMajor=False):
    self._points = points
    self._data = data
    self.shaper = shaper 
    self.shaper.colMajor = colMajor ##HACK: This modification is bad form...would rather do a copy

  ##TODO: Add the ability to get points m...n 
  def points(self):
    """Returns the set of [x,y,w,h] points for this glyphset.
       Access to raw data is through _points.
    """

    if (type(self.shaper) is Literals):
      if type(self._points) is list:
        return np.array(self._points, order="F")
      elif type(self._points) is np.ndarray:
        return self._points
      else:
        ValueError("Unhandled (literal) points type: %s" % type(self._points))
    else:
      ##TODO: Setup the shaper utilities to go directly to fortran order...save a copy 
      return np.array(self.shaper(self._points), order="F")

  def data(self): 
    return self._data
  
  def bounds(self):
    """Compute bounds of the glyph-set.  Returns (X,Y,W,H)
    
       Assumes a 'simple' layout where the min and max input values
       determine the min and max positional values.
    """
    minX=float("inf")
    maxX=float("-inf")
    minY=float("inf")
    maxY=float("-inf")
    for g in self.points():
      (x,y,w,h) = g

      minX=min(minX, x)
      maxX=max(maxX, x+w)
      minY=min(minY, y)
      maxY=max(maxY, y+h)
    return (minX, minY, maxX-minX, maxY-minY) 
#    points = self.points();
#    minX=points[:,0].min()
#    maxX=points[:,0].max()
#    minY=points[:,1].min()
#    maxY=points[:,1].max()
#    maxW=points[:,2].max()
#    maxH=points[:,3].max()
#
#    shapes = self.shaper([[minX,maxX],[minY,maxY],[maxW,maxW],[maxH,maxH]])
#    minX = shapes[0][0]
#    minY = shapes[0][1]
#    maxX = shapes[1][0]
#    maxY = shapes[1][1]
#    width = shapes[1][2]
#    height = shapes[1][3]
#
#    return (minX, minY, maxX+width, maxY+height) 
  


### Shapers.....
class Shaper(object):
  fns = None #List of functions to apply 
  code = None
  colMajor = False 
  
  ##TODO: When getting subsets of the data out of glyphset.points(), remove this colMajor and handle it up in glyphset instead
  def __call__(self, vals):
    if not self.colMajor:
      return [map(lambda f: f(val), self.fns) for val in vals]
    else:
      return [map(lambda f: f(val), self.fns) for val in zip(*vals)]

class Literals(Shaper):
  """Optimization marker, tested in Glyphset and conversions skipped if present.
     Using this class asserts that the _poitns value of the Glyphset is the set
     of points to feed into a geometry renderer, and thus no projection from
     data space to canvas space is required.
     Use with caution...
  """
  def __init__(self, code):
    self.code = code

  def __call__(self, vals):
    return vals

class ToRect(Shaper):
  """Creates rectangles using functions to build x,y,w,h."""
  code = ShapeCodes.RECT
  def __init__(self, tox, toy, tow, toh):
    self.fns = [tox,toy,tow,toh]

class ToLine(Shaper):
  """Creates lines using functions to build x1,y1,x2,y2"""
  code = ShapeCodes.LINE
  def __init__(self, tox1, toy1, tox2, toy2):
    self.fns = [tox1, toy1, tox2, toy2]

class ToPoint(Shaper):
  """Create a single point, using functions to build x,y,0,0"""
  code = ShapeCodes.POINT
  def __init__(self, tox, toy, tow, toh):
    self.fns = [tox, toy,lambda(x): 0,lambda(x): 0]

#### Utilities for shapers....
def const(v):
  """Create a function that always returns a specific value.
  
    * v -- The value the returned function always returns
  """
  def f(a): return v
  return f

def idx(i):
  """Create a function that expects an indexable thing and returns a value at the passed index
  
  * i -- The index that will be used 
  """
  def f(a): return a[i]
  return f



import core
import numpy as np
import math

try:
  from numba import autojit
except ImportError:
  print "Error loading numba."
  autojit = lambda f: f


######## Aggregators ########
class Count(core.Aggregator):
  """Count the number of items that fall into a particular grid element."""
  out_type=np.int32
  identity=0

  def allocate(self, width, height, glyphset, infos):
    return np.zeros((width, height), dtype=self.out_type)

  def combine(self, existing, points, shapecode, val):
    update = core.glyphAggregates(points, shapecode, 1, self.identity)  
    existing[points[0]:points[2],points[1]:points[3]] += update

  def rollup(*vals):
    return reduce(lambda a,b: a+b,  vals)



class Sum(core.Aggregator):
  """Count the number of items that fall into a particular grid element."""
  out_type=np.int32
  identity=0

  def allocate(self, width, height, glyphset, infos):
    return np.zeros((width, height), dtype=self.out_type)

  def combine(self, existing, points, shapecode, val):
    update = core.glyphAggregates(points, shapecode, val, self.identity)  
    existing[points[0]:points[2],points[1]:points[3]] += update

  def rollup(*vals):
    return reduce(lambda a,b: a+b,  vals)


######## Shaders ##########
class FlattenCategories(core.Shader):
  """Convert a set of category-counts into just a set of counts"""
  out_type=(1,np.int32)
  in_type=("A",np.int32) #A is for "any", all cells must be the same size, but the exact size doesn't matter

  def shade(self, grid):
    return grid.sum(axis=1)


class Floor(core.Shader):
  def shade(self, grid):
    return np.floor(grid)
 
class Interpolate(core.Shader):
  """Interpolate between two numbers.
     Projects the input values between the low and high values passed.
     The Default is 0 to 1.
     Empty values are preserved (default is np.nan).
  """
  def __init__(self, low=0, high=1, empty=np.nan):
    self.low = low
    self.high = high
    self.empty = empty

  def shade(self, grid):
    mask = (grid == self.empty)
    min = grid[~mask].min()
    max = grid[~mask].max()
    span = float(max-min) 
    percents = (grid-min)/span
    return percents * (self.high-self.low)

class Power(core.Shader):
  """Raise to a power.  Power may be fracional."""
  def __init__(self, pow):
    self.pow = pow

  def shade(self, grid):
    return np.power(grid, self.pow)


class Cuberoot(Power):
  def __init__(self): 
    super(Cuberoot, self).__init__(1/3.0)

class Sqrt(core.Shader):
  def shade(self, grid): 
    return np.sqrt(grid, self.pow)

class Spread(core.PixelShader):
  """Spreads the values out in a regular pattern.
  
     TODO: Currently only does square spread.  Extend to other shapes.
     TODO: Restricted to numbers right now...implement corresponding thing for categories...might be 'generic'
  """

  def __init__(self, size):
    self.size = size

  def makegrid(self, grid):
    return np.zeros_like(grid)

  def pixelfunc(self, grid, x, y):
    minx = max(0, x-math.floor(self.size/2.0))
    maxx = x+math.ceil(self.size/2.0)
    miny = max(0,y-math.floor(self.size/2.0))
    maxy = y+math.ceil(self.size/2.0)

    parts = grid[minx:maxx, miny:maxy]
    return parts.sum()



class AbsSegment(core.Shader):
  """
  Paint all pixels with aggregate value above divider one color 
  and below the divider another.
  """
  in_type=(1,np.number)
  out_type=(4,np.int32)

  def __init__(self, low, high, divider):
    self.high = high
    self.low = low
    self.divider = float(divider)

  def shade(self, grid):
    (width, height) = grid.shape[0], grid.shape[1]
    outgrid = np.ndarray((width, height, 4), dtype=np.uint8)
    mask = (grid >= self.divider) 
    outgrid[mask] = self.high
    outgrid[~mask] = self.low
    return outgrid

class InterpolateColors(core.Shader):
  """
  High-definition interpolation between two colors.
  Zero-values are treated separately from other values.
 
  * low -- Color ot use for lowest value
  * high -- Color to use for highest values 
  * log -- Set to desired log base to use log-based interpolation 
           (use True or "e" for base-e; default is False)
  * reserve -- color to use for empty cells
  """
  in_type=(1,np.number)
  out_type=(4,np.int32)

  def __init__(self, low, high, log=False, reserve=core.Color(255,255,255,255), empty=np.nan):
    self.low=low
    self.high=high
    self.reserve=reserve
    self.log=log
    self.empty=empty

  
  ##TODO: there are issues with zeros here....
  def _log(self,  grid):
    mask = (grid == self.empty)
    min = grid[~mask].min()
    max = grid[~mask].max()
    
    grid[mask] = 1
    if (self.log==10):
      min = math.log10(min)
      max = math.log10(max)
      span = float(max-min)
      percents = (np.log10(grid)-min)/span
    elif (self.log == math.e or self.log == True):
      min = math.log(min)
      max = math.log(max)
      span = float(max-min)
      percents = (np.log(grid)-min)/span
    elif (self.log==2):
      min = math.log(min, self.log)
      max = math.log(max, self.log)
      span = float(max-min)
      percents = (np.log2(grid)-min)/span
    else:
      rebase = math.log(self.log)
      min = math.log(min, self.log)
      max = math.log(max, self.log)
      span = float(max-min)
      percents = ((np.log(grid)/rebase)-min)/span
    
    grid[mask] = 0
    
    colorspan = self.high.asarray().astype(np.int32) - self.low.asarray().astype(np.int32)

    outgrid = (percents[:,:,np.newaxis] * colorspan[np.newaxis,np.newaxis,:] + self.low.asarray()).astype(np.uint8)
    outgrid[mask] = self.reserve
                 
    return outgrid


  def _linear(self, grid):
    mask = (grid == self.empty)
    min = grid[~mask].min()
    max = grid[~mask].max()
    span = float(max-min) 
    percents = (grid-min)/span
    
    colorspan = self.high.asarray().astype(np.int32) - self.low.asarray().astype(np.int32)
    outgrid = (percents[:,:,np.newaxis] * colorspan[np.newaxis,np.newaxis,:] + self.low.asarray()).astype(np.uint8)
    outgrid[mask] = self.reserve
    return outgrid

    
  def shade(self, grid):
    if (self.log):
      return self._log(grid)
    else :
      return self._linear(grid)

    



""" 
Tools for working with counts from multiple categories of data at once.

Categories are modeled as stakced 2D arrays.  Each category is in its
own slice of the stack.
"""

import core 
import numpy as np
from math import log

try:
  from numba import autojit
except ImportError:
  print "Error loading numba."
  autojit = lambda f: f

##### Aggregator ##########
class CountCategories(core.Aggregator):
  """Count the number of items that fall into a particular grid element.
  """
  out_type = np.int32
  identity=np.asarray([0])
  cats=None

  def allocate(self, width, height, glyphset, infos):
    self.cats = np.unique(infos)
    return np.zeros((width, height, len(self.cats)), dtype=self.out_type)

  def combine(self, existing, points, shapecode, val):
    entry = np.zeros(self.cats.shape[0])
    idx = np.nonzero(self.cats==val)[0][0]
    entry[idx] = 1
    update = core.glyphAggregates(points, shapecode, entry, self.identity)  
    existing[points[0]:points[2],points[1]:points[3]] += update

  def rollup(*vals):
    return reduce(lambda x,y: x+y,  vals)



##### Shaders ########
class ToCounts(core.Shader):
  """Convert from count-by-categories to just raw counts.
     Then data shader functions from the count module can be used.
  """
  @staticmethod
  def shade(grid, dtype=np.int32):
    return np.sum(grid, axis=2, dtype=dtype)

class Select(core.Shader):
  """Get the counts from just one category.

     Operates by taking a single plane of the count of categories.

     TODO: Consider changing shade to take a wrapper 'grid' that can carry info
           like a category-label-to-grid-slice mapping....
  """
  
  def __init__(self, slice):
    """slice -- Which slice of the aggregates grid should be returned"""
    self.slice = slice

  def shade(aggregates, dtype=np.int32):
    return aggregates[:,:,slice]



class MinPercent(core.Shader):
  """If the item in the specified bin represents more than a certain percent
     of the total number of items, color it as "above" otherwise, color as "below"
       
     TODO: Change from idx to category label, lookup idx for 'cat' parameter
  
     * cutoff -- percent value to split above and below coloring
     * cat -- integer indicating which category number to use  
     * above  -- color to paint above (default a red)
     * below  -- color to paint below (default a blue)
     * background -- color to paint when there are no values (default is clear)
  """
     
  def __init__(self, 
               cutoff, 
               cat=0,
               above=core.Color(228, 26, 28,255), 
               below=core.Color(55, 126, 184,255), 
               background=core.Color(255,255,255,0)):

    self.cutoff = cutoff
    self.cat = cat  
    self.above = above.asarray()
    self.below = below.asarray()
    self.background = background.asarray()

  def shade(self, grid):
    (width, height) = grid.shape[0], grid.shape[1]
    outgrid = np.empty((width, height, 4), dtype=np.uint8)
    
    sums = ToCounts.shade(grid, dtype=np.float32)
    maskbg = sums == 0 
    mask = (grid[:,:,self.cat]/sums) >= self.cutoff

    outgrid[mask] = self.above
    outgrid[~mask] = self.below
    outgrid[maskbg] = self.background
    return outgrid

class HDAlpha(core.Shader):
  def __init__(self, colors, 
               background=core.Color(255,255,255,255), alphamin=0, log=False, logbase=10):
    """colors -- a list of colors in cateogry-order.
                 TODO: Change to a dictionary of category-to-color mapping
       alphamin -- minimum alpha value when (default is 0)
       log -- Log-based interpolate? (default is false)
       logbase -- Base to use if log is true (default is 10)
       background -- Color when the category list is empty (default is white)
    """
    #self.colors = dict(zip(colors.keys(), map(lambda v: v.asarray(), colors.values)))
    self.catcolors = np.array(map(lambda v: v.asarray(), colors))
    self.background = background.asarray()
    self.alphamin = alphamin
    self.log = log
    self.logbase = logbase

  def shade(self, grid):
    sums = ToCounts.shade(grid, dtype=np.float32)
    mask = (sums!=0)

    colors = opaqueblend(self.catcolors, grid, sums)
    colors[~mask] = self.background
    alpha(colors, sums, mask, self.alphamin, self.log, self.logbase)
    return colors



##### Utilities #######
def alpha(colors, sums, mask, alphamin, dolog=False, base=10):
  maxval = sums.max()

  if (dolog):
    base = log(base)
    maxval = log(maxval)/base
    sums[mask] = log(sums[mask])/base

  np.putmask(colors[:,:,3], mask,((alphamin + ((1-alphamin) * (sums/maxval)))*255).astype(np.uint8))

def opaqueblend(catcolors, counts, sums):
  weighted = (counts/sums[:,:,np.newaxis]).astype(float)
  weighted = catcolors[np.newaxis,np.newaxis,:]*weighted[:,:,:,np.newaxis]
  colors = weighted.sum(axis=2).astype(np.uint8)
  return colors 



from __future__ import print_function
from distutils.core import setup, Extension
from os.path import abspath, exists, join, dirname
import sys
import os 
import sys

### Flag to indicate that libdispatch should be used (OS X only)
DISPATCH_FLAG = "--dispatch"

if DISPATCH_FLAG in sys.argv:
  transform = Extension('abstract_rendering.transform_libdispatch',
                        ['abstract_rendering/transform_libdispatch.cpp'], 
                        extra_compile_args=['-std=c++11','-O3', '-Wall', '-march=native', '-fno-rtti', '-fno-exceptions', '-fPIC', '-lstdc++'])
  del sys.argv[sys.argv.index(DISPATCH_FLAG)]

else:
  transform = Extension('abstract_rendering.transform',
                        ['abstract_rendering/transform.cpp'],
                        extra_compile_args=['-std=c++11','-O3', '-Wall', '-march=native', '-fno-rtti', '-fno-exceptions', '-fPIC', '-lstdc++'])
def getsitepackages():
    """Returns a list containing all global site-packages directories
    (and possibly site-python)."""

    _is_64bit = (getattr(sys, 'maxsize', None) or getattr(sys, 'maxint')) > 2**32
    _is_pypy = hasattr(sys, 'pypy_version_info')
    _is_jython = sys.platform[:4] == 'java'

    prefixes = [sys.prefix, sys.exec_prefix]

    sitepackages = []
    seen = set()

    for prefix in prefixes:
        if not prefix or prefix in seen:
            continue
        seen.add(prefix)

        if sys.platform in ('os2emx', 'riscos') or _is_jython:
            sitedirs = [os.path.join(prefix, "Lib", "site-packages")]
        elif _is_pypy:
            sitedirs = [os.path.join(prefix, 'site-packages')]
        elif sys.platform == 'darwin' and prefix == sys.prefix:
            if prefix.startswith("/System/Library/Frameworks/"): # Apple's Python
                sitedirs = [os.path.join("/Library/Python", sys.version[:3], "site-packages"),
                            os.path.join(prefix, "Extras", "lib", "python")]

            else:  # any other Python distros on OSX work this way
                sitedirs = [os.path.join(prefix, "lib",
                            "python" + sys.version[:3], "site-packages")]

        elif os.sep == '/':
            sitedirs = [os.path.join(prefix,
                                     "lib",
                                     "python" + sys.version[:3],
                                     "site-packages"),
                        os.path.join(prefix, "lib", "site-python"),
                        os.path.join(prefix, "python" + sys.version[:3], "lib-dynload")]
            lib64_dir = os.path.join(prefix, "lib64", "python" + sys.version[:3], "site-packages")
            if (os.path.exists(lib64_dir) and
                os.path.realpath(lib64_dir) not in [os.path.realpath(p) for p in sitedirs]):
                if _is_64bit:
                    sitedirs.insert(0, lib64_dir)
                else:
                    sitedirs.append(lib64_dir)
            try:
                # sys.getobjects only available in --with-pydebug build
                sys.getobjects
                sitedirs.insert(0, os.path.join(sitedirs[0], 'debug'))
            except AttributeError:
                pass
            # Debian-specific dist-packages directories:
            if sys.version[0] == '2':
                sitedirs.append(os.path.join(prefix, "lib",
                                             "python" + sys.version[:3],
                                             "dist-packages"))
            else:
                sitedirs.append(os.path.join(prefix, "lib",
                                             "python" + sys.version[0],
                                             "dist-packages"))
            sitedirs.append(os.path.join(prefix, "local/lib",
                                         "python" + sys.version[:3],
                                         "dist-packages"))
            sitedirs.append(os.path.join(prefix, "lib", "dist-python"))
        else:
            sitedirs = [prefix, os.path.join(prefix, "lib", "site-packages")]
        if sys.platform == 'darwin':
            # for framework builds *only* we add the standard Apple
            # locations. Currently only per-user, but /Library and
            # /Network/Library could be added too
            if 'Python.framework' in prefix:
                home = os.environ.get('HOME')
                if home:
                    sitedirs.append(
                        os.path.join(home,
                                     'Library',
                                     'Python',
                                     sys.version[:3],
                                     'site-packages'))
        for sitedir in sitedirs:
            sitepackages.append(os.path.abspath(sitedir))
    return sitepackages



site_packages = getsitepackages()[0]
old_dir = join(site_packages, "abstract_rendering")
path_file = join(site_packages, "abstract_rendering.pth")
path = abspath(dirname(__file__))

if 'develop' in sys.argv:
    print("Develop mode.")
    if os.path.isdir(old_dir):
      print("  - Removing package %s." % old_dir)
      import shutil
      shutil.rmtree(old_dir)
    with open(path_file, "w+") as f:
        f.write(path)
    print("  - writing path '%s' to %s" % (path, path_file))
    print()
    sys.exit()

setup(name='abstract_rendering', 
      version='0.1.0',
      description='Rendering as a binning process',
      author='Joseph Cottam',
      author_email='jcottam@indiana.edu',
      url='https://github.com/JosephCottam/AbstractRendering',
      package_dir = {'abstract_rendering' : 'abstract_rendering'},
      py_modules=['abstract_rendering.core', 
                  'abstract_rendering.general', 
                  'abstract_rendering.categories',
                  'abstract_rendering.fast_project',
                  'abstract_rendering.geometry',
                  'abstract_rendering.glyphset', 
                  'abstract_rendering.infos',
                  'abstract_rendering.numeric'],
      ext_modules=[transform]
     )




"""Each info function returns a function that takes two arguments:
     shape and data.  The 'shape' is the projected shape information.
     The 'data' is the data package associated with that information.
"""


def const(v):
  """Return the value passed.""" 
  def f(glyph, data):
    return v
  return f

def val(default=None):
  """Return the entire data value.  This is the info-version of 'id'.
  """
  def f(glyph, data):
    if (data is None) :
      return default
    else:
      return data
  return f 

def valAt(i, default=None):
  """Return the value at a given index in the data part of the input.
     On error returns the default value.
  """
  def f(glyph, data):
    try :
      return data[i] 
    except:
      return default
  return f 

def attribute(att, default=None):
  """"Return the value under a given attribute in the data part of hte input."""
  def f(glyph, data):
    try :
      return getattr(data, att)
    except:
      return default
  return f 
